
import functools
import traceback
import sys
from multiprocessing import Value, Lock
from synapseclient.core.utils import printTransferProgress


def notifyMe(syn, messageSubject='', retries=0):
    """Function decorator that notifies you via email whenever an function completes running or there is a failure.

    :param syn:             A synapse object as obtained with syn = synapseclient.login()
    :param messageSubject:  A string with subject line for sent out messages.
    :param retries:         Number of retries to attempt on failure (default=0)

    Example::

        # to decorate a function that you define
        from synapseutils import notifyMe
        import synapseclient
        syn = synapseclient.login()

        @notifyMe(syn, 'Long running function', retries=2)
        def my_function(x):
            doing_something()
            return long_runtime_func(x)

        my_function(123)

        #############################
        # to wrap a function that already exists
        from synapseutils import notifyMe
        import synapseclient
        syn = synapseclient.login()

        notify_decorator = notifyMe(syn, 'Long running query', retries=2)
        my_query = notify_decorator(syn.tableQuery)
        results = my_query("select id from syn1223")

        #############################
    """
    def notify_decorator(func):
        @functools.wraps(func)
        def with_retry_and_messaging(*args, **kwargs):
            attempt = 0
            destination = syn.getUserProfile()['ownerId']
            while attempt <= retries:
                try:
                    output = func(*args, **kwargs)
                    syn.sendMessage([destination],  messageSubject,
                                    messageBody='Call to %s completed successfully!' % func.__name__)
                    return output
                except Exception as e:
                    sys.stderr.write(traceback.format_exc())
                    syn.sendMessage([destination], messageSubject,
                                    messageBody=('Encountered a temporary Failure during upload.  '
                                                 'Will retry %i more times. \n\n Error message was:\n%s\n\n%s'
                                                 % (retries-attempt, e, traceback.format_exc())))
                    attempt += 1
        return with_retry_and_messaging
    return notify_decorator


def with_progress_bar(func, totalCalls, prefix='', postfix='', isBytes=False):
    """Wraps a function to add a progress bar based on the number of calls to that function.

    :param func:        Function being wrapped with progress Bar
    :param totalCalls:  total number of items/bytes when completed
    :param prefix:      String printed before progress bar
    :param prefix:      String printed after progress bar
    :param isBytes:     A boolean indicating weather to convert bytes to kB, MB, GB etc.

    :return: a wrapped function that contains a progress bar
    """
    completed = Value('d', 0)
    lock = Lock()

    def progress(*args, **kwargs):
        with lock:
            completed.value += 1
        printTransferProgress(completed.value, totalCalls, prefix, postfix, isBytes)
        return func(*args, **kwargs)
    return progress



from .download_threads import (
    DownloadRequest,
    download_file,
    shared_executor,
    SYNAPSE_DEFAULT_DOWNLOAD_PART_SIZE,
)

__all__ = ['DownloadRequest', 'download_file', 'shared_executor', 'SYNAPSE_DEFAULT_DOWNLOAD_PART_SIZE']



"""
************************
Synapse Multipart Upload
************************

Implements the client side of `Synapse multipart upload`_, which provides a
robust means of uploading large files (into the 10s of GB). End users should
not need to call any of these functions directly.

.. _Synapse multipart upload:
 http://docs.synapse.org/rest/index.html#org.sagebionetworks.file.controller.UploadController

"""

import concurrent.futures
from contextlib import contextmanager
import hashlib
import json
import math
import mimetypes
import os
import re
import requests
import threading
import time
from typing import List, Mapping

from synapseclient.core.retry import with_retry
from synapseclient.core import pool_provider
from synapseclient.core.constants import concrete_types
from synapseclient.core.exceptions import (
    _raise_for_status,  # why is is this a single underscore
    SynapseHTTPError,
    SynapseUploadAbortedException,
    SynapseUploadFailedException,
)
from synapseclient.core.utils import md5_for_file, MB, Spinner

# AWS limits
MAX_NUMBER_OF_PARTS = 10000
MIN_PART_SIZE = 5 * MB

# ancient tribal knowledge
DEFAULT_PART_SIZE = 8 * MB
MAX_RETRIES = 7


_thread_local = threading.local()


@contextmanager
def shared_executor(executor):
    """An outside process that will eventually trigger an upload through the this module
    can configure a shared Executor by running its code within this context manager."""
    _thread_local.executor = executor
    try:
        yield
    finally:
        del _thread_local.executor


@contextmanager
def _executor(max_threads, shutdown_wait):
    """Yields an executor for running some asynchronous code, either obtaining the executor
    from the shared_executor or otherwise creating one.

    :param max_threads: the maxmimum number of threads a created executor should use
    :param shutdown_wait: whether a created executor should shutdown after running the yielded to code
    """
    executor = getattr(_thread_local, 'executor', None)
    shutdown_after = False
    if not executor:
        shutdown_after = True
        executor = pool_provider.get_executor(thread_count=max_threads)

    try:
        yield executor
    finally:
        if shutdown_after:
            executor.shutdown(wait=shutdown_wait)


class UploadAttempt:

    def __init__(
        self,
        syn,
        dest_file_name,

        upload_request_payload,
        part_request_body_provider_fn,
        md5_fn,

        max_threads: int,
        force_restart: bool,
    ):
        self._syn = syn
        self._dest_file_name = dest_file_name
        self._part_size = upload_request_payload['partSizeBytes']

        self._upload_request_payload = upload_request_payload

        self._part_request_body_provider_fn = part_request_body_provider_fn
        self._md5_fn = md5_fn

        self._max_threads = max_threads
        self._force_restart = force_restart

        self._lock = threading.Lock()
        self._aborted = False

        # populated later
        self._upload_id: str = None
        self._pre_signed_part_urls: Mapping[int, str] = None

    @classmethod
    def _get_remaining_part_numbers(cls, upload_status):
        part_numbers = []
        parts_state = upload_status['partsState']

        # parts are 1-based
        for i, part_status in enumerate(parts_state, 1):
            if part_status == '0':
                part_numbers.append(i)

        return len(parts_state), part_numbers

    @classmethod
    def _get_thread_session(cls):
        # get a lazily initialized requests.Session from the thread.
        # we want to share a requests.Session over the course of a thread
        # to take advantage of persistent http connection. we put it on a
        # thread local rather that in the task closure since a connection can
        # be reused across separate part uploads so no reason to restrict it
        # per worker task.
        session = getattr(_thread_local, 'session', None)
        if not session:
            session = _thread_local.session = requests.Session()
        return session

    def _is_copy(self):
        # is this a copy or upload request
        return self._upload_request_payload.get('concreteType') == concrete_types.MULTIPART_UPLOAD_COPY_REQUEST

    def _create_synapse_upload(self):
        return self._syn.restPOST(
            "/file/multipart?forceRestart={}".format(
                str(self._force_restart).lower()
            ),
            json.dumps(self._upload_request_payload),
            endpoint=self._syn.fileHandleEndpoint,
        )

    def _fetch_pre_signed_part_urls(
        self,
        upload_id: str,
        part_numbers: List[int],
        requests_session: requests.Session = None,
    ) -> Mapping[int, str]:

        uri = "/file/multipart/{upload_id}/presigned/url/batch".format(
            upload_id=upload_id
        )
        body = {
            'uploadId': upload_id,
            'partNumbers': part_numbers,
        }

        response = self._syn.restPOST(
            uri,
            json.dumps(body),
            requests_session=requests_session,
            endpoint=self._syn.fileHandleEndpoint,
        )

        part_urls = {}
        for part in response['partPresignedUrls']:
            part_urls[part['partNumber']] = (
                part['uploadPresignedUrl'],
                part.get('signedHeaders', {}),
            )

        return part_urls

    def _refresh_pre_signed_part_urls(
        self,
        part_number: int,
        expired_url: str,
    ):
        """Refresh all unfetched presigned urls, and return the refreshed
        url for the given part number. If an existing expired_url is passed
        and the url for the given part has already changed that new url
        will be returned without a refresh (i.e. it is assumed that another
        thread has already refreshed the url since the passed url expired).

        :param part_number: the part number whose refreshed url should
            be returned
        :param expired_url: the url that was detected as expired triggering
            this refresh

        """
        with self._lock:
            current_url = self._pre_signed_part_urls[part_number]
            if current_url != expired_url:
                # if the url has already changed since the given url
                # was detected as expired we can assume that another
                # thread already refreshed the url and can avoid the extra
                # fetch.
                refreshed_url = current_url
            else:
                self._pre_signed_part_urls = self._fetch_pre_signed_part_urls(
                    self._upload_id,
                    list(self._pre_signed_part_urls.keys()),
                )

                refreshed_url = self._pre_signed_part_urls[part_number]

            return refreshed_url

    def _handle_part(self, part_number):
        with self._lock:
            if self._aborted:
                # this upload attempt has already been aborted
                # so we short circuit the attempt to upload this part
                raise SynapseUploadAbortedException(
                    "Upload aborted, skipping part {}".format(part_number)
                )

            part_url, signed_headers = self._pre_signed_part_urls.get(part_number)

        session = self._get_thread_session()

        # obtain the body (i.e. the upload bytes) for the given part number.
        body = self._part_request_body_provider_fn(part_number) if self._part_request_body_provider_fn else None
        part_size = len(body) if body else 0
        for retry in range(2):
            def put_fn():
                return session.put(part_url, body, headers=signed_headers)
            try:
                # use our backoff mechanism here, we have encountered 500s on puts to AWS signed urls
                response = with_retry(put_fn, retry_exceptions=[requests.exceptions.ConnectionError])
                _raise_for_status(response)

                # completed upload part to s3 successfully
                break

            except SynapseHTTPError as ex:
                if ex.response.status_code == 403 and retry < 1:
                    # we interpret this to mean our pre_signed url expired.
                    self._syn.logger.debug(
                        "The pre-signed upload URL for part {} has expired."
                        "Refreshing urls and retrying.\n".format(part_number)
                    )

                    # we refresh all the urls and obtain this part's
                    # specific url for the retry
                    part_url, signed_headers = self._refresh_pre_signed_part_urls(
                        part_number,
                        part_url,
                    )

                else:
                    raise

        md5_hex = self._md5_fn(body, response)

        # now tell synapse that we uploaded that part successfully
        self._syn.restPUT(
            "/file/multipart/{upload_id}/add/{part_number}?partMD5Hex={md5}"
            .format(
                upload_id=self._upload_id,
                part_number=part_number,
                md5=md5_hex,
            ),
            requests_session=session,
            endpoint=self._syn.fileHandleEndpoint
        )

        # remove so future batch pre_signed url fetches will exclude this part
        with self._lock:
            del self._pre_signed_part_urls[part_number]

        return part_number, part_size

    def _upload_parts(self, part_count, remaining_part_numbers):
        time_upload_started = time.time()
        completed_part_count = part_count - len(remaining_part_numbers)
        file_size = self._upload_request_payload.get('fileSizeBytes')

        if not self._is_copy():
            # we won't have bytes to measure during a copy so the byte oriented progress bar is not useful
            progress = previously_transferred = min(
                completed_part_count * self._part_size,
                file_size,
            )

            self._syn._print_transfer_progress(
                progress,
                file_size,
                prefix='Uploading',
                postfix=self._dest_file_name,
                previouslyTransferred=previously_transferred,
            )

        self._pre_signed_part_urls = self._fetch_pre_signed_part_urls(
            self._upload_id,
            remaining_part_numbers,
        )

        futures = []
        with _executor(self._max_threads, False) as executor:
            # we don't wait on the shutdown since we do so ourselves below

            for part_number in remaining_part_numbers:
                futures.append(
                    executor.submit(
                        self._handle_part,
                        part_number,
                    )
                )

        for result in concurrent.futures.as_completed(futures):
            try:
                _, part_size = result.result()

                if part_size and not self._is_copy():
                    progress += part_size
                    self._syn._print_transfer_progress(
                        min(progress, file_size),
                        file_size,
                        prefix='Uploading',
                        postfix=self._dest_file_name,
                        dt=time.time() - time_upload_started,
                        previouslyTransferred=previously_transferred,
                    )
            except (Exception, KeyboardInterrupt) as cause:
                with self._lock:
                    self._aborted = True

                # wait for all threads to complete before
                # raising the exception, we don't want to return
                # control while there are still threads from this
                # upload attempt running
                concurrent.futures.wait(futures)

                if isinstance(cause, KeyboardInterrupt):
                    raise SynapseUploadAbortedException(
                        "User interrupted upload"
                    )

                raise SynapseUploadFailedException(
                    "Part upload failed"
                ) from cause

    def _complete_upload(self):
        upload_status_response = self._syn.restPUT(
            "/file/multipart/{upload_id}/complete".format(
                upload_id=self._upload_id,
            ),
            requests_session=self._get_thread_session(),
            endpoint=self._syn.fileHandleEndpoint,
        )

        upload_state = upload_status_response.get('state')
        if upload_state != 'COMPLETED':
            # at this point we think successfully uploaded all the parts
            # but the upload status isn't complete, we'll throw an error
            # and let a subsequent attempt try to reconcile
            raise SynapseUploadFailedException(
                "Upload status has an unexpected state {}".format(upload_state)
            )

        return upload_status_response

    def __call__(self):
        upload_status_response = self._create_synapse_upload()
        upload_state = upload_status_response.get('state')

        if upload_state != 'COMPLETED':
            self._upload_id = upload_status_response['uploadId']
            part_count, remaining_part_numbers =\
                self._get_remaining_part_numbers(
                    upload_status_response
                )

            # if no remaining part numbers then all the parts have been
            # uploaded but the upload has not been marked complete.
            if remaining_part_numbers:
                self._upload_parts(part_count, remaining_part_numbers)
            upload_status_response = self._complete_upload()

        return upload_status_response


def _get_file_chunk(file_path, part_number, chunk_size):
    """
    Read the nth chunk from the file.
    """
    with open(file_path, 'rb') as f:
        f.seek((part_number - 1) * chunk_size)
        return f.read(chunk_size)


def _get_data_chunk(data, part_number, chunk_size):
    """
    Return the nth chunk of a buffer.
    """
    return data[(part_number - 1) * chunk_size: part_number * chunk_size]


def _get_part_size(part_size, file_size):
    part_size = part_size or DEFAULT_PART_SIZE

    # can't exceed the maximum allowed num parts
    part_size = max(
        part_size,
        MIN_PART_SIZE,
        int(math.ceil(file_size / MAX_NUMBER_OF_PARTS))
    )
    return part_size


def multipart_upload_file(
    syn,
    file_path: str,
    dest_file_name: str = None,
    content_type: str = None,
    part_size: int = None,
    storage_location_id: str = None,
    preview: bool = True,
    force_restart: bool = False,
    max_threads: int = None,
) -> str:
    """
    Upload a file to a Synapse upload destination in chunks.

    :param syn:                 a Synapse object
    :param file_path:           the file to upload
    :param dest_file_name:      upload as a different filename
    :param content_type:        contentType`_
    :param part_size:           number of bytes per part. Minimum 5MB.
    :param storage_location_id: an id indicating where the file should be
                                stored. Retrieved from Synapse's
                                UploadDestination
    :param preview:             True to generate a preview
    :param force_restart        True to restart a previously initiated upload
                                from scratch, False to try to resume
    :param max_threads          number of concurrent threads to devote
                                to upload

    :return: a File Handle ID

    Keyword arguments are passed down to
    :py:func:`_multipart_upload` and :py:func:`_start_multipart_upload`.

    .. _contentType:
     https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17
    """

    if not os.path.exists(file_path):
        raise IOError('File "{}" not found.'.format(file_path))
    if os.path.isdir(file_path):

        raise IOError('File "{}" is a directory.'.format(file_path))

    file_size = os.path.getsize(file_path)
    if not dest_file_name:
        dest_file_name = os.path.basename(file_path)

    if content_type is None:
        mime_type, _ = mimetypes.guess_type(file_path, strict=False)
        content_type = mime_type or 'application/octet-stream'

    callback_func = Spinner().print_tick if not syn.silent else None
    md5_hex = md5_for_file(file_path, callback=callback_func).hexdigest()

    part_size = _get_part_size(part_size, file_size)

    upload_request = {
        'concreteType': concrete_types.MULTIPART_UPLOAD_REQUEST,
        'contentType': content_type,
        'contentMD5Hex': md5_hex,
        'fileName': dest_file_name,
        'fileSizeBytes': file_size,
        'generatePreview': preview,
        'partSizeBytes': part_size,
        'storageLocationId': storage_location_id,
    }

    def part_fn(part_number):
        return _get_file_chunk(file_path, part_number, part_size)

    def md5_fn(part, _):
        md5 = hashlib.md5()
        md5.update(part)
        return md5.hexdigest()

    return _multipart_upload(
        syn,
        dest_file_name,

        upload_request,
        part_fn,
        md5_fn,

        force_restart=force_restart,
        max_threads=max_threads,
    )


def multipart_upload_string(
    syn,
    text: str,
    dest_file_name: str = None,
    part_size: int = None,
    content_type: str = None,
    storage_location_id: str = None,
    preview: bool = True,
    force_restart: bool = False,
    max_threads: int = None,
):
    """
    Upload a file to a Synapse upload destination in chunks.

    :param syn:                 a Synapse object
    :param text:                a string to upload as a file.
    :param dest_file_name:      upload as a different filename
    :param content_type:        contentType`_
    :param part_size:           number of bytes per part. Minimum 5MB.
    :param storage_location_id: an id indicating where the file should be
                                stored. Retrieved from Synapse's
                                UploadDestination
    :param preview:             True to generate a preview
    :param force_restart        True to restart a previously initiated upload
                                from scratch, False to try to resume
    :param max_threads          number of concurrent threads to devote
                                to upload

    :return: a File Handle ID

    Keyword arguments are passed down to
    :py:func:`_multipart_upload` and :py:func:`_start_multipart_upload`.

    .. _contentType:
     https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17
    """

    data = text.encode('utf-8')
    file_size = len(data)
    md5_hex = hashlib.md5(data).hexdigest()

    if not dest_file_name:
        dest_file_name = 'message.txt'

    if not content_type:
        content_type = "text/plain; charset=utf-8"

    part_size = _get_part_size(part_size, file_size)

    upload_request = {
        'concreteType': concrete_types.MULTIPART_UPLOAD_REQUEST,
        'contentType': content_type,
        'contentMD5Hex': md5_hex,
        'fileName': dest_file_name,
        'fileSizeBytes': file_size,
        'generatePreview': preview,
        'partSizeBytes': part_size,
        'storageLocationId': storage_location_id,
    }

    def part_fn(part_number):
        return _get_data_chunk(data, part_number, part_size)

    def md5_fn(part, _):
        md5 = hashlib.md5()
        md5.update(part)
        return md5.hexdigest()

    part_size = _get_part_size(part_size, file_size)
    return _multipart_upload(
        syn,
        dest_file_name,

        upload_request,
        part_fn,
        md5_fn,

        force_restart=force_restart,
        max_threads=max_threads,
    )


def multipart_copy(
    syn,
    source_file_handle_association,
    dest_file_name: str = None,
    part_size: int = None,
    storage_location_id: str = None,
    preview: bool = True,
    force_restart: bool = False,
    max_threads: int = None,
):

    part_size = part_size or DEFAULT_PART_SIZE

    upload_request = {
        'concreteType': concrete_types.MULTIPART_UPLOAD_COPY_REQUEST,
        'fileName': dest_file_name,
        'generatePreview': preview,
        'partSizeBytes': part_size,
        'sourceFileHandleAssociation': source_file_handle_association,
        'storageLocationId': storage_location_id,
    }

    def part_request_body_provider_fn(part_num):
        # for an upload copy there are no bytes
        return None

    def md5_fn(_, response):
        # for a multipart copy we use the md5 returned by the UploadPartCopy command
        # when we add the part to the Synapse upload

        # we extract the md5 from the <ETag> element in the response.
        # use lookahead and lookbehind to find the opening and closing ETag elements but
        # do not include those in the match, thus the entire matched string (group 0) will be
        # what was between those elements.
        md5_hex = re.search('(?<=<ETag>).*?(?=<\\/ETag>)', (response.content.decode('utf-8'))).group(0)

        # remove quotes found in the ETag to get at the normalized ETag
        return md5_hex.replace('&quot;', '').replace('"', '')

    return _multipart_upload(
        syn,
        dest_file_name,

        upload_request,
        part_request_body_provider_fn,
        md5_fn,

        force_restart=force_restart,
        max_threads=max_threads,
    )


def _multipart_upload(
    syn,
    dest_file_name,

    upload_request,
    part_fn,
    md5_fn,

    force_restart: bool = False,
    max_threads: int = None,
):

    if max_threads is None:
        max_threads = pool_provider.DEFAULT_NUM_THREADS

    max_threads = max(max_threads, 1)

    retry = 0
    while True:
        try:
            upload_status_response = UploadAttempt(
                syn,
                dest_file_name,

                upload_request,
                part_fn,
                md5_fn,

                max_threads,

                # only force_restart the first time through (if requested).
                # a retry after a caught exception will not restart the upload
                # from scratch.
                force_restart and retry == 0,
            )()

            # success
            return upload_status_response['resultFileHandleId']

        except SynapseUploadFailedException:
            if retry < MAX_RETRIES:
                retry += 1
            else:
                raise



try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading
import datetime

import concurrent.futures
from contextlib import contextmanager
from http import HTTPStatus
import os
from requests import Session, Response
from requests.adapters import HTTPAdapter
from typing import Generator, NamedTuple
from urllib.parse import urlparse, parse_qs
from urllib3.util.retry import Retry
import time

from synapseclient.core.exceptions import SynapseError
from synapseclient.core.pool_provider import get_executor
from synapseclient.core.retry import (
    with_retry,
    RETRYABLE_CONNECTION_ERRORS,
    RETRYABLE_CONNECTION_EXCEPTIONS,
)

# constants
MAX_QUEUE_SIZE: int = 20
MiB: int = 2 ** 20
SYNAPSE_DEFAULT_DOWNLOAD_PART_SIZE: int = 8 * MiB
ISO_AWS_STR_FORMAT: str = '%Y%m%dT%H%M%SZ'
CONNECT_FACTOR: int = 3
BACK_OFF_FACTOR: float = 0.5


_thread_local = _threading.local()


@contextmanager
def shared_executor(executor):
    """An outside process that will eventually trigger a download through the this module
    can configure a shared Executor by running its code within this context manager."""
    _thread_local.executor = executor
    try:
        yield
    finally:
        del _thread_local.executor


class DownloadRequest(NamedTuple):
    """
    A request to download a file from Synapse

    ...

    Attributes
    ----------
    file_handle_id : int
        The file handle ID to download.
    object_id : str
        The Synapse object this file associated to.
    object_type : str
        the type of the associated Synapse object.
    path : str
        The local path to download the file to.
        This path can be either absolute path or relative path from where the code is executed to the download location.
    """
    file_handle_id: int
    object_id: str
    object_type: str
    path: str


class TransferStatus(object):
    """
    Transfer progress parameters. Lock should be acquired via `with trasfer_status:` before accessing attributes
    Attributes
    ----------
    total_bytes_to_be_transferred: int
    transferred: int
    """
    total_bytes_to_be_transferred: int
    transferred: int

    def __init__(self, total_bytes_to_be_transferred: int):
        self.total_bytes_to_be_transferred = total_bytes_to_be_transferred
        self.transferred = 0
        self._t0 = time.time()

    def elapsed_time(self) -> float:
        """
        :return: time since this object was created (assuming same time as transfer started)
        """
        return time.time() - self._t0


class PresignedUrlInfo(NamedTuple):
    """
    Information about a retrieved presigned-url

    ...

    Attributes
    ----------
    file_name: str
        name of the file for the presigned url
    url: str
        the actual presigned url
    expiration_utc: datetime.datetime
        datetime in UTC at which the url will expire
    """
    file_name: str
    url: str
    expiration_utc: datetime.datetime


class PresignedUrlProvider(object):
    """
    Provides an un-exipired pre-signed url to download a file
    """
    request: DownloadRequest
    _cached_info: PresignedUrlInfo

    # offset parameter used to buffer url expiration checks, time in seconds
    _TIME_BUFFER: datetime.timedelta = datetime.timedelta(seconds=5)

    def __init__(self, client, request: DownloadRequest):
        self.client = client
        self.request: DownloadRequest = request
        self._cached_info: PresignedUrlInfo = self._get_pre_signed_info()
        self._lock = _threading.Lock()

    def get_info(self) -> PresignedUrlInfo:
        with self._lock:
            if datetime.datetime.utcnow() + PresignedUrlProvider._TIME_BUFFER >= self._cached_info.expiration_utc:
                self._cached_info = self._get_pre_signed_info()

            return self._cached_info

    def _get_pre_signed_info(self) -> PresignedUrlInfo:
        """
        Returns the file_name and pre-signed url for download as specified in request

        :return: PresignedUrlInfo
        """
        # noinspection PyProtectedMember
        response = self.client._getFileHandleDownload(
            self.request.file_handle_id,
            self.request.object_id,
            objectType=self.request.object_type,
        )
        file_name = response["fileHandle"]["fileName"]
        pre_signed_url = response["preSignedURL"]
        return PresignedUrlInfo(file_name, pre_signed_url, _pre_signed_url_expiration_time(pre_signed_url))


def _generate_chunk_ranges(file_size: int,
                           ) -> Generator:
    """
    Creates a generator which yields byte ranges and meta data required to make a range request download of url and
    write the data to file_name located at path. Download chunk sizes are 8MB by default.

    :param file_size: The size of the file
    :return: A generator of byte ranges and meta data needed to download the file in a multi-threaded manner
    """
    for start in range(0, file_size, SYNAPSE_DEFAULT_DOWNLOAD_PART_SIZE):
        # the start and end of a range in HTTP are both inclusive
        end = min(start + SYNAPSE_DEFAULT_DOWNLOAD_PART_SIZE, file_size) - 1
        yield start, end


def _pre_signed_url_expiration_time(url: str) -> datetime:
    """
    Returns time at which a presigned url will expire

    :param url: A pre-signed download url from AWS
    :return: datetime in UTC of when the url will expire
    """
    parsed_query: dict = parse_qs(urlparse(url).query)
    time_made: str = parsed_query['X-Amz-Date'][0]
    time_made_datetime: datetime.datetime = datetime.datetime.strptime(time_made, ISO_AWS_STR_FORMAT)
    expires: str = parsed_query['X-Amz-Expires'][0]
    return time_made_datetime + datetime.timedelta(seconds=int(expires))


def _get_new_session() -> Session:
    """
    Creates a new requests.Session object with retry defined by CONNECT_FACTOR and BACK_OFF_FACTOR
    :return: A new requests.Session object
    """
    session = Session()
    retry = Retry(connect=CONNECT_FACTOR, backoff_factor=BACK_OFF_FACTOR)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def _get_file_size(url: str) -> int:
    """
    Gets the size of the file located at url
    :param url: The pre-signed url of the file
    :return: The size of the file in bytes
    """
    session = _get_new_session()
    res_get = session.get(url, stream=True)
    return int(res_get.headers['Content-Length'])


def download_file(
    client,
    download_request: DownloadRequest,
    *,
    max_concurrent_parts: int = None,
):
    """
    Main driver for the multi-threaded download. Users an ExecutorService, either set externally onto a thread
    local by an outside process, or creating one as needed otherwise.

    :param client: A synapseclient
    :param download_request: A batch of DownloadRequest objects specifying what Synapse files to download
    :param max_concurrent_parts: The maximum concurrent number parts to download at once when downloading this file
    """

    # we obtain an executor from a thread local if we are in the context of a Synapse sync
    # and wan't to re-use the same threadpool as was created for that
    executor = getattr(_thread_local, 'executor', None)
    shutdown_after = False
    if not executor:
        shutdown_after = True
        executor = get_executor(client.max_threads)

    max_concurrent_parts = max_concurrent_parts or client.max_threads
    try:
        downloader = _MultithreadedDownloader(client, executor, max_concurrent_parts)
        downloader.download_file(download_request)
    finally:
        # if we created the Executor for the purposes of processing this download we also
        # shut it down. if it was passed in from the outside then it's managed by the caller
        if shutdown_after:
            executor.shutdown()


def _get_thread_session():
    # get a lazily initialized requests.Session from the thread.
    # we want to share a requests.Session over the course of a thread
    # to take advantage of persistent http connection. we put it on a
    # thread local since Sessions are not thread safe so we need one per
    # active thread and since we're allowing the use of an externally provided
    # ExecutorService we don't can't really allocate a pool of Sessions ourselves
    session = getattr(_thread_local, 'session', None)
    if not session:
        session = _thread_local.session = _get_new_session()
    return session


class _MultithreadedDownloader:
    """
    An object to manage the downloading of a Synapse file in concurrent chunks from a URL
    that supports range headers.
    """

    def __init__(self, syn, executor, max_concurrent_parts):
        """
        :param syn:                     A synapseclient
        :param executor:                An ExecutorService that will be used to run part downloads in separate threads
        :param max_concurrent_parts:    An integer to specify the maximum number of concurrent parts that can be
                                        downloaded at once. If there are more parts than can be run concurrently
                                        they will be scheduled in the executor when previously running part downloads
                                        complete.
        """
        self._syn = syn
        self._executor = executor
        self._max_concurrent_parts = max_concurrent_parts

    def download_file(self, request):
        url_provider = PresignedUrlProvider(self._syn, request)

        url_info = url_provider.get_info()
        file_size = _get_file_size(url_info.url)
        chunk_range_generator = _generate_chunk_ranges(file_size)

        self._prep_file(request)

        transfer_status = TransferStatus(file_size)

        # the entrant thread runs in a loop doing the following:
        # 1. scheduling any additional part downloads as previous parts are completed
        # 2. writing any completed parts out to the local disk
        # 3. waiting for additional parts to complete
        pending_futures = set()
        completed_futures = set()
        try:
            while True:
                submitted_futures = self._submit_chunks(
                    url_provider,
                    chunk_range_generator,
                    pending_futures,
                )

                self._write_chunks(request, completed_futures, transfer_status)

                # once there is nothing else pending we are done with the file download
                pending_futures = pending_futures.union(submitted_futures)
                if not pending_futures:
                    break

                completed_futures, pending_futures = concurrent.futures.wait(
                    pending_futures,
                    return_when=concurrent.futures.FIRST_COMPLETED
                )

                self._check_for_errors(request, completed_futures)

        except BaseException:
            # on any exception (e.g. KeyboardInterrupt), attempt to cancel any pending futures.
            # if they are already running this won't have any effect though
            for future in pending_futures:
                future.cancel()

            try:
                os.remove(request.path)
            except FileNotFoundError:
                pass

            raise

    @staticmethod
    def _get_response_with_retry(presigned_url_provider, start: int, end: int) -> Response:
        session = _get_thread_session()
        range_header = {'Range': f'bytes={start}-{end}'}

        def session_get():
            return session.get(presigned_url_provider.get_info().url, headers=range_header)

        response = None
        cause = None
        try:
            # currently when doing a range request to AWS we retry on anything other than a 206.
            # this seems a bit excessive (i.e. some 400 statuses would suggest a non-retryable condition)
            # but for now matching previous behavior.
            response = with_retry(
                session_get,
                expected_status_codes=(HTTPStatus.PARTIAL_CONTENT,),
                retry_errors=RETRYABLE_CONNECTION_ERRORS,
                retry_exceptions=RETRYABLE_CONNECTION_EXCEPTIONS,
            )
        except Exception as ex:
            cause = ex

        if not response or response.status_code != HTTPStatus.PARTIAL_CONTENT:
            raise SynapseError(
                f'Could not download the file: {presigned_url_provider.get_info().file_name},'
                f' please try again.'
            ) from cause

        return start, response

    @staticmethod
    def _prep_file(request):
        # upon receiving the parts of the file we'll open the file
        # and write the specific byte ranges, but to open it in
        # r+ mode we need to to exist and be empty
        open(request.path, 'wb').close()

    def _submit_chunks(self, url_provider, chunk_range_generator, pending_futures):
        submit_count = self._max_concurrent_parts - len(pending_futures)
        submitted_futures = set()

        for chunk_range in chunk_range_generator:
            start, end = chunk_range
            chunk_future = self._executor.submit(
                self._get_response_with_retry,
                url_provider,
                start,
                end,
            )
            submitted_futures.add(chunk_future)

            if len(submitted_futures) == submit_count:
                break

        return submitted_futures

    def _write_chunks(self, request, completed_futures, transfer_status):
        if completed_futures:
            with open(request.path, 'rb+') as file_write:
                for chunk_future in completed_futures:
                    start, chunk_response = chunk_future.result()
                    chunk_data = chunk_response.content
                    file_write.seek(start)
                    file_write.write(chunk_response.content)

                    transfer_status.transferred += len(chunk_data)
                    self._syn._print_transfer_progress(transfer_status.transferred,
                                                       transfer_status.total_bytes_to_be_transferred,
                                                       'Downloading ', os.path.basename(request.path),
                                                       dt=transfer_status.elapsed_time())

    @staticmethod
    def _check_for_errors(request, completed_futures):
        # if any submitted part download failed we abort the download.
        # any retry/recovery should be attempted within the download method
        # submitted to the Executor, if an Exception was flagged on the Future
        # we consider it unrecoverable
        for completed_future in completed_futures:
            exception = completed_future.exception()
            if exception:
                raise ValueError(f"Failed downloading {request.object_id} to {request.path}") from exception



"""
This module provides access to thread pools to use in concurrent work.
Both a multiprocessing style ThreadPool and an Executor based pool
are available, Executors should be preferred for new work as it provides
a more modern interface.

To use these wrappers for single thread environment, set the following:

    synapseclient.config.single_threaded = True
"""

from concurrent.futures import Executor, Future, ThreadPoolExecutor
import multiprocessing
import multiprocessing.dummy

from . import config

# +4 matches ThreadPoolExecutor, at least 5 threads for I/O bound tasks
# https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
DEFAULT_NUM_THREADS = multiprocessing.cpu_count() + 4


class SingleThreadPool:

    def map(self, func, iterable):
        for item in iterable:
            func(item)

    def terminate(self):
        pass


class SingleThreadExecutor(Executor):
    def __init__(self):
        self._shutdown = False

    def submit(self, fn, *args, **kwargs):
        if self._shutdown:
            raise RuntimeError('cannot schedule new futures after shutdown')

        f = Future()
        try:
            result = fn(*args, **kwargs)
        except BaseException as e:
            f.set_exception(e)
        else:
            f.set_result(result)

        return f


class FakeLock:

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        pass


class SingleValue:

    value = None

    def __init__(self, type, value):
        self.value = value

    def get_lock(self):
        return FakeLock()


def get_pool():
    if config.single_threaded:
        return SingleThreadPool()
    else:
        return multiprocessing.dummy.Pool(DEFAULT_NUM_THREADS)


def get_executor(thread_count=DEFAULT_NUM_THREADS):
    """
    Provides an Executor as defined by the client config suitable
    for running tasks work as defined by the client config.

    :param thread_count:        number of concurrent threads

    :return: an Executor
    """
    if config.single_threaded:
        return SingleThreadExecutor()
    else:
        return ThreadPoolExecutor(max_workers=thread_count)


def get_value(type, value):
    if config.single_threaded:
        return SingleValue(type, value)
    else:
        return multiprocessing.Value(type, value)



from synapseclient.entity import is_container
import os


def walk(syn, synId):
    """
    Traverse through the hierarchy of files and folders stored under the synId. Has the same behavior as os.walk()

    :param syn:     A synapse object: syn = synapseclient.login()- Must be logged into synapse

    :param synId:   A synapse ID of a folder or project

    Example::

        walkedPath = walk(syn, "syn1234")

        for dirpath, dirname, filename in walkedPath:
            print(dirpath)
            print(dirname) #All the folders in the directory path
            print(filename) #All the files in the directory path

    """
    return _helpWalk(syn, synId)


# Helper function to hide the newpath parameter
def _helpWalk(syn, synId, newpath=None):
    starting = syn.get(synId, downloadFile=False)
    # If the first file is not a container, return immediately
    if newpath is None and not is_container(starting):
        return
    elif newpath is None:
        dirpath = (starting.name, synId)
    else:
        dirpath = (newpath, synId)
    dirs = []
    nondirs = []
    results = syn.getChildren(synId)
    for i in results:
        if is_container(i):
            dirs.append((i['name'], i['id']))
        else:
            nondirs.append((i['name'], i['id']))
    yield dirpath, dirs, nondirs
    for name in dirs:
        newpath = os.path.join(dirpath[0], name[0])
        for x in _helpWalk(syn, name[1], newpath=newpath):
            yield x



"""
*****************
Version Functions
*****************

Check for latest version and recommend upgrade::

    synapseclient.check_for_updates()

Print release notes for installed version of client::

    synapseclient.release_notes()

.. automethod:: synapseclient.version_check.check_for_updates
.. automethod:: synapseclient.version_check.release_notes

"""

import json
import pkg_resources
import re
import requests
import synapseclient
import sys


_VERSION_URL = 'https://raw.githubusercontent.com/Sage-Bionetworks/synapsePythonClient/master/synapseclient/synapsePythonClient'  # noqa


def version_check(current_version=None, version_url=_VERSION_URL, check_for_point_releases=False):
    """
    Gets the latest version information from version_url and check against the current version.
    Recommends upgrade, if a newer version exists.

    :returns: True if current version is the latest release (or higher) version,
              False otherwise.
    """

    try:
        if not current_version:
            current_version = synapseclient.__version__

        version_info = _get_version_info(version_url)

        current_base_version = _strip_dev_suffix(current_version)

        # Check blacklist
        if current_base_version in version_info['blacklist'] or current_version in version_info['blacklist']:
            msg = ("\nPLEASE UPGRADE YOUR CLIENT\n\nUpgrading your SynapseClient is required. "
                   "Please upgrade your client by typing:\n"
                   "    pip install --upgrade synapseclient\n\n")
            raise SystemExit(msg)

        if 'message' in version_info:
            sys.stderr.write(version_info['message'] + '\n')

        levels = 3 if check_for_point_releases else 2

        # Compare with latest version
        if _version_tuple(current_version, levels=levels) < _version_tuple(version_info['latestVersion'],
                                                                           levels=levels):
            sys.stderr.write("\nUPGRADE AVAILABLE\n\nA more recent version of the Synapse Client (%s) "
                             "is available. Your version (%s) can be upgraded by typing:\n"
                             "    pip install --upgrade synapseclient\n\n" %
                             (version_info['latestVersion'], current_version,))
            if 'releaseNotes' in version_info:
                sys.stderr.write('Python Synapse Client version %s release notes\n\n'
                                 % version_info['latestVersion'])
                sys.stderr.write(version_info['releaseNotes'] + '\n\n')
            return False

    except Exception as e:
        # Don't prevent the client from running if something goes wrong
        sys.stderr.write("Exception in version check: %s\n" % (str(e),))
        return False

    return True


def check_for_updates():
    """
    Check for the existence of newer versions of the client, reporting both current release version and development
    version.

    For help installing development versions of the client, see the docs for
    :py:mod:`synapseclient` or the `README.md <https://github.com/Sage-Bionetworks/synapsePythonClient>`_.
    """
    sys.stderr.write('Python Synapse Client\n')
    sys.stderr.write('currently running version:  %s\n' % synapseclient.__version__)

    release_version_info = _get_version_info(_VERSION_URL)
    sys.stderr.write('latest release version:     %s\n' % release_version_info['latestVersion'])

    if _version_tuple(synapseclient.__version__, levels=3) < _version_tuple(release_version_info['latestVersion'],
                                                                            levels=3):
        print(("\nUPGRADE AVAILABLE\n\nA more recent version of the Synapse Client (%s) is available. "
               "Your version (%s) can be upgraded by typing:\n"
               "    pip install --upgrade synapseclient\n\n") % (release_version_info['latestVersion'],
                                                                 synapseclient.__version__,))
    else:
        sys.stderr.write('\nYour Synapse client is up to date!\n')


def release_notes(version_url=None):
    """
    Print release notes for the installed version of the client or latest release or development version if version_url
    is supplied.

    :param version_url: Defaults to None, meaning release notes for the installed version. Alternatives are:
                            - synapseclient.version_check._VERSION_URL
                            - synapseclient.version_check._DEV_VERSION_URL
    """
    version_info = _get_version_info(version_url)
    sys.stderr.write('Python Synapse Client version %s release notes\n\n' % version_info['latestVersion'])
    if 'releaseNotes' in version_info:
        sys.stderr.write(version_info['releaseNotes'] + '\n')


def _strip_dev_suffix(version):
    return re.sub(r'\.dev\d+', '', version)


def _version_tuple(version, levels=2):
    """
    Take a version number as a string delimited by periods and return a tuple with the desired number of levels.
    For example::

        print(version_tuple('0.5.1.dev1', levels=2))
        ('0', '5')
    """
    v = _strip_dev_suffix(version).split('.')
    v = v[0:min(len(v), levels)]
    if len(v) < levels:
        v = v + ['0'] * (levels-len(v))
    return tuple(v)


def _get_version_info(version_url=_VERSION_URL):
    if version_url is None:
        return json.loads(pkg_resources.resource_string('synapseclient', 'synapsePythonClient').decode())
    else:
        headers = {'Accept': 'application/json; charset=UTF-8'}
        headers.update(synapseclient.USER_AGENT)
        return requests.get(version_url, headers=headers).json()


# If this file is run as a script, print current version
# then perform version check
if __name__ == "__main__":
    print("Version check")
    print("=============")
    print("Python Synapse Client version %s" % synapseclient.__version__)

    print("Check against production version:")
    if version_check():
        print("ok")

    print("Check against local copy of version file:")
    if version_check(version_url=None):
        print("ok")



"""
*****************
Utility Functions
*****************

Utility functions useful in the implementation and testing of the Synapse client.

"""
import base64
import cgi
import collections.abc
import datetime
import errno
import functools
import hashlib
import importlib
import inspect
import numbers
import os
import platform
import random
import re
import requests
import sys
import tempfile
import threading
import urllib.parse as urllib_parse
import uuid
import warnings


UNIX_EPOCH = datetime.datetime(1970, 1, 1, 0, 0)
ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.000Z"
ISO_FORMAT_MICROS = "%Y-%m-%dT%H:%M:%S.%fZ"
GB = 2**30
MB = 2**20
KB = 2**10
BUFFER_SIZE = 8*KB


def md5_for_file(filename, block_size=2 * MB, callback=None):
    """
    Calculates the MD5 of the given file.
    See `source <http://stackoverflow.com/questions/1131220/get-md5-hash-of-a-files-without-open-it-in-python>`_.

    :param filename:   The file to read in
    :param block_size: How much of the file to read in at once (bytes).
                       Defaults to 2 MB
    :param callback:   The callback function that help us show loading spinner on terminal.
                       Defaults to None
    :returns: The MD5
    """

    md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        while True:
            if callback:
                callback()
            data = f.read(block_size)
            if not data:
                break
            md5.update(data)
    return md5


def download_file(url, localFilepath=None):
    """
    Downloads a remote file.

    :param localFilePath: May be None, in which case a temporary file is created

    :returns: localFilePath
    """

    f = None
    try:
        if localFilepath:
            dir = os.path.dirname(localFilepath)
            if not os.path.exists(dir):
                os.makedirs(dir)
            f = open(localFilepath, 'wb')
        else:
            f = tempfile.NamedTemporaryFile(delete=False)
            localFilepath = f.name

        r = requests.get(url, stream=True)
        toBeTransferred = float(r.headers['content-length'])
        for nChunks, chunk in enumerate(r.iter_content(chunk_size=1024*10)):
            if chunk:
                f.write(chunk)
                printTransferProgress(nChunks*1024*10, toBeTransferred)
    finally:
        if f:
            f.close()
            printTransferProgress(toBeTransferred, toBeTransferred)

    return localFilepath


def extract_filename(content_disposition_header, default_filename=None):
    """
    Extract a filename from an HTTP content-disposition header field.

    See `this memo <http://tools.ietf.org/html/rfc6266>`_ and `this package <http://pypi.python.org/pypi/rfc6266>`_
    for cryptic details.
    """

    if not content_disposition_header:
        return default_filename
    value, params = cgi.parse_header(content_disposition_header)
    return params.get('filename', default_filename)


def extract_user_name(profile):
    """
    Extract a displayable user name from a user's profile
    """
    if 'userName' in profile and profile['userName']:
        return profile['userName']
    elif 'displayName' in profile and profile['displayName']:
        return profile['displayName']
    else:
        if 'firstName' in profile and profile['firstName'] and 'lastName' in profile and profile['lastName']:
            return profile['firstName'] + ' ' + profile['lastName']
        elif 'lastName' in profile and profile['lastName']:
            return profile['lastName']
        elif 'firstName' in profile and profile['firstName']:
            return profile['firstName']
        else:
            return str(profile.get('id', 'Unknown-user'))


def _get_from_members_items_or_properties(obj, key):
    if hasattr(obj, key):
        return getattr(obj, key)
    try:
        if hasattr(obj, 'properties') and key in obj.properties:
            return obj.properties[key]
        if key in obj:
            return obj[key]
        else:
            return obj['properties'][key]
    except (KeyError, TypeError):
        # We cannot get the key from this obj. So this case will be treated as key not found.
        pass
    return None


# TODO: what does this do on an unsaved Synapse Entity object?
def id_of(obj):
    """
    Try to figure out the Synapse ID of the given object.

    :param obj: May be a string, Entity object, or dictionary

    :returns: The ID or throws an exception
    """
    if isinstance(obj, str):
        return str(obj)
    if isinstance(obj, numbers.Number):
        return str(obj)

    id_attr_names = ['id', 'ownerId', 'tableId']  # possible attribute names for a synapse Id
    for attribute_name in id_attr_names:
        syn_id = _get_from_members_items_or_properties(obj, attribute_name)
        if syn_id is not None:
            return str(syn_id)

    raise ValueError('Invalid parameters: couldn\'t find id of ' + str(obj))


def concrete_type_of(obj: collections.abc.Mapping):
    """
    Return the concrete type of an object representing a Synapse entity.
    This is meant to operate either against an actual Entity object, or the lighter
    weight dictionary returned by Synapse#getChildren, both of which are Mappings.
    """
    concrete_type = None
    if isinstance(obj, collections.abc.Mapping):
        for key in ('concreteType', 'type'):
            concrete_type = obj.get(key)
            if concrete_type:
                break

    if not isinstance(concrete_type, str) or not concrete_type.startswith('org.sagebionetworks.repo.model'):
        raise ValueError('Unable to determine concreteType')

    return concrete_type


def is_in_path(id, path):
    """Determines whether id is in the path as returned from /entity/{id}/path

    :param id:      synapse id string
    :param path:    object as returned from '/entity/{id}/path'

    :returns: True or False
    """
    return id in [item['id'] for item in path['path']]


def get_properties(entity):
    """Returns the dictionary of properties of the given Entity."""

    return entity.properties if hasattr(entity, 'properties') else entity


def is_url(s):
    """Return True if the string appears to be a valid URL."""
    if isinstance(s, str):
        try:
            url_parts = urllib_parse.urlsplit(s)
            # looks like a Windows drive letter?
            if len(url_parts.scheme) == 1 and url_parts.scheme.isalpha():
                return False
            if url_parts.scheme == 'file' and bool(url_parts.path):
                return True
            return bool(url_parts.scheme) and bool(url_parts.netloc)
        except Exception:
            return False
    return False


def as_url(s):
    """Tries to convert the input into a proper URL."""
    url_parts = urllib_parse.urlsplit(s)
    # Windows drive letter?
    if len(url_parts.scheme) == 1 and url_parts.scheme.isalpha():
        return 'file:///%s' % str(s).replace("\\", "/")
    if url_parts.scheme:
        return url_parts.geturl()
    else:
        return 'file://%s' % str(s)


def guess_file_name(string):
    """Tries to derive a filename from an arbitrary string."""
    path = normalize_path(urllib_parse.urlparse(string).path)
    tokens = [x for x in path.split('/') if x != '']
    if len(tokens) > 0:
        return tokens[-1]

    # Try scrubbing the path of illegal characters
    if len(path) > 0:
        path = re.sub(r"[^a-zA-Z0-9_.+() -]", "", path)
    if len(path) > 0:
        return path
    raise ValueError("Could not derive a name from %s" % string)


def normalize_path(path):
    """Transforms a path into an absolute path with forward slashes only."""
    if path is None:
        return None
    return re.sub(r'\\', '/', os.path.normcase(os.path.abspath(path)))


def equal_paths(path1, path2):
    """
    Compare file paths in a platform neutral way
    """
    return normalize_path(path1) == normalize_path(path2)


def file_url_to_path(url, verify_exists=False):
    """
    Convert a file URL to a path, handling some odd cases around Windows paths.

    :param url:             a file URL
    :param verify_exists:   If true, return an populated dict only if the resulting file path exists on the local file
                            system.

    :returns: a path or None if the URL is not a file URL.
    """
    parts = urllib_parse.urlsplit(url)
    if parts.scheme == 'file' or parts.scheme == '':
        path = parts.path
        # A windows file URL, for example file:///c:/WINDOWS/asdf.txt
        # will get back a path of: /c:/WINDOWS/asdf.txt, which we need to fix by
        # lopping off the leading slash character. Apparently, the Python developers
        # think this is not a bug: http://bugs.python.org/issue7965
        if re.match(r'\/[A-Za-z]:', path):
            path = path[1:]
        if os.path.exists(path) or not verify_exists:
            return path
    return None


def is_same_base_url(url1, url2):
    """Compares two urls to see if they are the same excluding up to the base path

    :param url1: a URL
    :param url2: a second URL

    :returns: Boolean
    """
    url1 = urllib_parse.urlsplit(url1)
    url2 = urllib_parse.urlsplit(url2)
    return (url1.scheme == url2.scheme and
            url1.hostname == url2.hostname)


def is_synapse_id(obj):
    """If the input is a Synapse ID return it, otherwise return None"""
    if isinstance(obj, str):
        m = re.match(r'(syn\d+)', obj)
        if m:
            return m.group(1)
    return None


def is_date(dt):
    """Objects of class datetime.date and datetime.datetime will be recognized as dates"""
    return isinstance(dt, datetime.date) or isinstance(dt, datetime.datetime)


def to_list(value):
    """Convert the value (an iterable or a scalar value) to a list."""
    if isinstance(value, collections.abc.Iterable) and not isinstance(value, str):
        return list(value)
    else:
        return [value]


def _to_iterable(value):
    """Convert the value (an iterable or a scalar value) to an iterable."""
    if isinstance(value, collections.abc.Iterable):
        return value
    return value,


def make_bogus_data_file(n=100, seed=None):
    """
    Makes a bogus data file for testing. It is the caller's responsibility to clean up the file when finished.

    :param n:    How many random floating point numbers to be written into the file, separated by commas
    :param seed: Random seed for the random numbers

    :returns: The name of the file
    """

    if seed is not None:
        random.seed(seed)
    data = [random.gauss(mu=0.0, sigma=1.0) for i in range(n)]

    f = tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False)
    try:
        f.write(", ".join(str(n) for n in data))
        f.write("\n")
    finally:
        f.close()

    return normalize_path(f.name)


def make_bogus_binary_file(n=1*MB, filepath=None, printprogress=False):
    """
    Makes a bogus binary data file for testing. It is the caller's responsibility to clean up the file when finished.

    :param n:       How many bytes to write

    :returns: The name of the file
    """

    with open(filepath, 'wb') if filepath else tempfile.NamedTemporaryFile(mode='wb', suffix=".dat", delete=False) as f:
        if not filepath:
            filepath = f.name
        progress = 0
        remaining = n
        while remaining > 0:
            buff_size = int(min(remaining, 1*MB))
            f.write(os.urandom(buff_size))
            remaining -= buff_size
            if printprogress:
                progress += buff_size
                printTransferProgress(progress, n, 'Generated ', filepath)
        return normalize_path(filepath)


def to_unix_epoch_time(dt):
    """
    Convert either `datetime.date or datetime.datetime objects <http://docs.python.org/2/library/datetime.html>`_
    to UNIX time.
    """

    if type(dt) == datetime.date:
        return (dt - UNIX_EPOCH.date()).total_seconds() * 1000
    return int((dt - UNIX_EPOCH).total_seconds() * 1000)


def to_unix_epoch_time_secs(dt):
    """
    Convert either `datetime.date or datetime.datetime objects <http://docs.python.org/2/library/datetime.html>`_
    to UNIX time.
    """

    if type(dt) == datetime.date:
        return (dt - UNIX_EPOCH.date()).total_seconds()
    return (dt - UNIX_EPOCH).total_seconds()


def from_unix_epoch_time_secs(secs):
    """Returns a Datetime object given milliseconds since midnight Jan 1, 1970."""
    if isinstance(secs, str):
        secs = float(secs)

    # utcfromtimestamp() fails for negative values (dates before 1970-1-1) on Windows
    # so, here's a hack that enables ancient events, such as Chris's birthday to be
    # converted from milliseconds since the UNIX epoch to higher level Datetime objects. Ha!
    if platform.system() == 'Windows' and secs < 0:
        mirror_date = datetime.datetime.utcfromtimestamp(abs(secs))
        return UNIX_EPOCH - (mirror_date - UNIX_EPOCH)
    return datetime.datetime.utcfromtimestamp(secs)


def from_unix_epoch_time(ms):
    """Returns a Datetime object given milliseconds since midnight Jan 1, 1970."""

    if isinstance(ms, str):
        ms = float(ms)
    return from_unix_epoch_time_secs(ms/1000.0)


def datetime_to_iso(dt, sep="T"):
    # Round microseconds to milliseconds (as expected by older clients)
    # and add back the "Z" at the end.
    # see: http://stackoverflow.com/questions/30266188/how-to-convert-date-string-to-iso8601-standard
    fmt = "{time.year:04}-{time.month:02}-{time.day:02}" \
          "{sep}{time.hour:02}:{time.minute:02}:{time.second:02}.{millisecond:03}{tz}"
    if dt.microsecond >= 999500:
        dt -= datetime.timedelta(microseconds=dt.microsecond)
        dt += datetime.timedelta(seconds=1)
    return fmt.format(time=dt, millisecond=int(round(dt.microsecond/1000.0)), tz="Z", sep=sep)


def iso_to_datetime(iso_time):
    return datetime.datetime.strptime(iso_time, ISO_FORMAT_MICROS)


def format_time_interval(seconds):
    """Format a time interval given in seconds to a readable value, e.g. \"5 minutes, 37 seconds\"."""

    periods = (
        ('year',        60*60*24*365),
        ('month',       60*60*24*30),
        ('day',         60*60*24),
        ('hour',        60*60),
        ('minute',      60),
        ('second',      1),)

    result = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds or period_name == 'second':
            period_value, seconds = divmod(seconds, period_seconds)
            if period_value > 0 or period_name == 'second':
                if period_value == 1:
                    result.append("%d %s" % (period_value, period_name))
                else:
                    result.append("%d %ss" % (period_value, period_name))
    return ", ".join(result)


def _find_used(activity, predicate):
    """Finds a particular used resource in an activity that matches a predicate."""

    for resource in activity['used']:
        if predicate(resource):
            return resource
    return None


def itersubclasses(cls, _seen=None):
    """
    http://code.activestate.com/recipes/576949/ (r3)

    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>>
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]
    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None:
        _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for inner_sub in itersubclasses(sub, _seen):
                yield inner_sub


def normalize_whitespace(s):
    """
    Strips the string and replace all whitespace sequences and other non-printable characters with a single space.
    """
    assert isinstance(s, str)
    return re.sub(r'[\x00-\x20\s]+', ' ', s.strip())


def normalize_lines(s):
    assert isinstance(s, str)
    s2 = re.sub(r'[\t ]*\n[\t ]*', '\n', s.strip())
    return re.sub(r'[\t ]+', ' ', s2)


def _synapse_error_msg(ex):
    """
    Format a human readable error message
    """
    if isinstance(ex, str):
        return ex

    # one line for the root exception and then an additional line per chained cause
    error_str = ""
    depth = 0
    while ex:
        error_str += \
            '\n' + ("  " * depth) + ("caused by " if depth > 0 else "") + ex.__class__.__name__ + ': ' + str(ex)

        ex = ex.__cause__
        if ex:
            depth += 1
        else:
            break

    return error_str + '\n\n'


def _limit_and_offset(uri, limit=None, offset=None):
    """
    Set limit and/or offset query parameters of the given URI.
    """
    parts = urllib_parse.urlparse(uri)
    query = urllib_parse.parse_qs(parts.query)
    if limit is None:
        query.pop('limit', None)
    else:
        query['limit'] = limit
    if offset is None:
        query.pop('offset', None)
    else:
        query['offset'] = offset

    new_query_string = urllib_parse.urlencode(query, doseq=True)
    return urllib_parse.urlunparse(urllib_parse.ParseResult(
        scheme=parts.scheme,
        netloc=parts.netloc,
        path=parts.path,
        params=parts.params,
        query=new_query_string,
        fragment=parts.fragment))


def query_limit_and_offset(query, hard_limit=1000):
    """
    Extract limit and offset from the end of a query string.

    :returns:   A triple containing the query with limit and offset removed, the limit at most equal to the hard_limit,
                and the offset which
                defaults to 1
    """
    # Regex a lower-case string to simplify matching
    tempQueryStr = query.lower()
    regex = r'\A(.*\s)(offset|limit)\s*(\d*\s*)\Z'

    # Continue to strip off and save the last limit/offset
    match = re.search(regex, tempQueryStr)
    options = {}
    while match is not None:
        options[match.group(2)] = int(match.group(3))
        tempQueryStr = match.group(1)
        match = re.search(regex, tempQueryStr)

    # Get a truncated version of the original query string (not in lower-case)
    query = query[:len(tempQueryStr)].strip()

    # Continue querying until the entire query has been fetched (or crash out)
    limit = min(options.get('limit', hard_limit), hard_limit)
    offset = options.get('offset', 1)

    return query, limit, offset


def extract_synapse_id_from_query(query):
    """
    An unfortunate hack to pull the synapse ID out of a table query of the form "select column1, column2 from syn12345
    where...." needed to build URLs for table services.
    """
    m = re.search(r"from\s+(syn\d+)", query, re.IGNORECASE)
    if m:
        return m.group(1)
    else:
        raise ValueError("Couldn't extract synapse ID from query: \"%s\"" % query)


# Derived from https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
def memoize(obj):
    cache = obj._memoize_cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        refresh = kwargs.pop('refresh', False)
        key = str(args) + str(kwargs)
        if refresh or key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


def printTransferProgress(transferred, toBeTransferred, prefix='', postfix='', isBytes=True, dt=None,
                          previouslyTransferred=0):
    """Prints a progress bar

    :param transferred:             a number of items/bytes completed
    :param toBeTransferred:         total number of items/bytes when completed
    :param prefix:                  String printed before progress bar
    :param postfix:                 String printed after progress bar
    :param isBytes:                 A boolean indicating whether to convert bytes to kB, MB, GB etc.
    :param dt:                      The time in seconds that has passed since transfer started is used to calculate rate
    :param previouslyTransferred:   the number of bytes that were already transferred before this transfer began
                                    (e.g. someone ctrl+c'd out of an upload and restarted it later)

    """
    if not sys.stdout.isatty():
        return
    barLength = 20  # Modify this to change the length of the progress bar
    status = ''
    rate = ''
    if dt is not None and dt != 0:
        rate = (transferred - previouslyTransferred)/float(dt)
        rate = '(%s/s)' % humanizeBytes(rate) if isBytes else rate
    if toBeTransferred < 0:
        defaultToBeTransferred = (barLength*1*MB)
        if transferred > defaultToBeTransferred:
            progress = float(transferred % defaultToBeTransferred) / defaultToBeTransferred
        else:
            progress = float(transferred) / defaultToBeTransferred
    elif toBeTransferred == 0:  # There is nothing to be transferred
        progress = 1
        status = "Done...\n"
    else:
        progress = float(transferred) / toBeTransferred
        if progress >= 1:
            progress = 1
            status = "Done...\n"
    block = int(round(barLength*progress))
    nbytes = humanizeBytes(transferred) if isBytes else transferred
    if toBeTransferred > 0:
        outOf = "/%s" % (humanizeBytes(toBeTransferred) if isBytes else toBeTransferred)
        percentage = "%4.2f%%" % (progress*100)
    else:
        outOf = ""
        percentage = ""
    text = "\r%s [%s]%s   %s%s %s %s %s    " % (prefix,
                                                "#"*block + "-"*(barLength-block),
                                                percentage,
                                                nbytes, outOf, rate,
                                                postfix, status)
    sys.stdout.write(text)
    sys.stdout.flush()


def humanizeBytes(num_bytes):
    if num_bytes is None:
        raise ValueError('bytes must be a number')

    num_bytes = float(num_bytes)
    units = ['bytes', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB']
    for i, unit in enumerate(units):
        if num_bytes < 1024:
            return '%3.1f%s' % (num_bytes, units[i])
        else:
            num_bytes /= 1024
    return 'Oops larger than Exabytes'


def touch(path, times=None):
    """
    Make sure a file exists. Update its access and modified times.
    """
    basedir = os.path.dirname(path)
    if not os.path.exists(basedir):
        try:
            os.makedirs(basedir)
        except OSError as err:
            # alternate processes might be creating these at the same time
            if err.errno != errno.EEXIST:
                raise

    with open(path, 'a'):
        os.utime(path, times)
    return path


def is_json(content_type):
    """detect if a content-type is JSON"""
    # The value of Content-Type defined here:
    # http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
    return content_type.lower().strip().startswith('application/json') if content_type else False


def find_data_file_handle(bundle):
    """Return the fileHandle whose ID matches the dataFileHandleId in an entity bundle"""
    for fileHandle in bundle['fileHandles']:
        if fileHandle['id'] == bundle['entity']['dataFileHandleId']:
            return fileHandle
    return None


def unique_filename(path):
    """Returns a unique path by appending (n) for some number n to the end of the filename."""

    base, ext = os.path.splitext(path)
    counter = 0
    while os.path.exists(path):
        counter += 1
        path = base + ("(%d)" % counter) + ext

    return path


class threadsafe_iter:
    """Takes an iterator/generator and makes it thread-safe by serializing call to the `next` method of given
    iterator/generator.
    See: http://anandology.com/blog/using-iterators-and-generators/
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.it)


def threadsafe_generator(f):
    """A decorator that takes a generator function and makes it thread-safe.
    See: http://anandology.com/blog/using-iterators-and-generators/
    """
    def g(*a, **kw):
        return threadsafe_iter(f(*a, **kw))
    return g


def extract_prefix(keys):
    """
    Takes a list of strings and extracts a common prefix delimited by a dot,
    for example::
        extract_prefix(["entity.bang", "entity.bar", "entity.bat"])
    entity.
    """
    prefixes = set()
    for key in keys:
        parts = key.split(".")
        if len(parts) > 1:
            prefixes.add(parts[0])
        else:
            return ""
    if len(prefixes) == 1:
        return prefixes.pop() + "."
    return ""


def temp_download_filename(destination, file_handle_id):
    suffix = "synapse_download_" + (str(file_handle_id)
                                    if file_handle_id
                                    else str(uuid.uuid4()))
    return os.path.join(destination, suffix) \
        if os.path.isdir(destination) \
        else destination + '.' + suffix


def extract_zip_file_to_directory(zip_file, zip_entry_name, target_dir):
    """
    Extracts a specified file in a zip to the specified directory
    :param zip_file:        an opened zip file. e.g. "with zipfile.ZipFile(zipfilepath) as zip_file:"
    :param zip_entry_name:  the name of the file to be extracted from the zip e.g. folderInsideZipIfAny/fileName.txt
    :param target_dir:      the directory to which the file will be extracted

    :return: full path to the extracted file
    """
    file_base_name = os.path.basename(zip_entry_name)  # base name of the file
    filepath = os.path.join(target_dir, file_base_name)  # file path to the cached file to write

    # Create the cache directory if it does not exist
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    # write the file from the zip into the cache
    with open(filepath, 'wb') as cache_file:
        cache_file.write(zip_file.read(zip_entry_name))

    return filepath


def is_integer(x):
    try:
        return float.is_integer(x)
    except TypeError:
        try:
            int(x)
            return True
        except (ValueError, TypeError):
            # anything that's not an integer, for example: empty string, None, 'NaN' or float('Nan')
            return False


def topolgical_sort(graph):
    """Given a graph in the form of a dictionary returns a sorted list

    Adapted from: http://blog.jupo.org/2012/04/06/topological-sorting-acyclic-directed-graphs/

    :param graph: a dictionary with values containing lists of keys referencing back into the dictionary

    :returns: sorted list of items
    """
    graph_unsorted = graph.copy()
    graph_sorted = []
    # Convert the unsorted graph into a hash table. This gives us
    # constant-time lookup for checking if edges are unresolved

    # Run until the unsorted graph is empty.
    while graph_unsorted:
        # Go through each of the node/edges pairs in the unsorted
        # graph. If a set of edges doesn't contain any nodes that
        # haven't been resolved, that is, that are still in the
        # unsorted graph, remove the pair from the unsorted graph,
        # and append it to the sorted graph. Note here that by using
        # using the items() method for iterating, a copy of the
        # unsorted graph is used, allowing us to modify the unsorted
        # graph as we move through it. We also keep a flag for
        # checking that that graph is acyclic, which is true if any
        # nodes are resolved during each pass through the graph. If
        # not, we need to bail out as the graph therefore can't be
        # sorted.
        acyclic = False
        for node, edges in list(graph_unsorted.items()):
            for edge in edges:
                if edge in graph_unsorted:
                    break
            else:
                acyclic = True
                del graph_unsorted[node]
                graph_sorted.append((node, edges))

        if not acyclic:
            # We've passed through all the unsorted nodes and
            # weren't able to resolve any of them, which means there
            # are nodes with cyclic edges that will never be resolved,
            # so we bail out with an error.
            raise RuntimeError("A cyclic dependency occurred."
                               " Some files in provenance reference each other circularly.")
    return graph_sorted


def caller_module_name(current_frame):
    """
    :param current_frame: use inspect.currentframe().
    :return: the name of the module calling the function, foo(), in which this calling_module() is invoked.
     Ignores callers that belong in the same module as foo()
    """

    current_frame_filename = current_frame.f_code.co_filename  # filename in which foo() resides

    # go back a frame takes us to the frame calling foo()
    caller_frame = current_frame.f_back
    caller_filename = caller_frame.f_code.co_filename

    # find the first frame that does not have the same filename. this ensures that we don't consider functions within
    # the same module as foo() that use foo() as a helper function
    while caller_filename == current_frame_filename:
        caller_frame = caller_frame.f_back
        caller_filename = caller_frame.f_code.co_filename

    return inspect.getmodulename(caller_filename)


def attempt_import(module_name, fail_message):

    try:
        return importlib.import_module(module_name)
    except ImportError:
        sys.stderr.write(
            (fail_message
             + "To install this library on Mac or Linux distributions:\n"
               "    (sudo) pip install %s\n\n"
               "On Windows, right click the Command Prompt(cmd.exe) and select 'Run as administrator' then:\n"
               "    pip install %s\n\n"
               "\n\n\n" % (module_name, module_name)))
        raise


def require_param(param, name):
    if param is None:
        raise ValueError("%s parameter is required." % name)


def snake_case(string):
    """Convert the given string from CamelCase to snake_case"""
    # https://stackoverflow.com/a/1176023
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


def is_base64_encoded(input_string):
    """Return whether the given input string appears to be base64 encoded"""
    if not input_string:
        # None, empty string are not considered encoded
        return False
    try:
        # see if we can decode it and then reencode it back to the input
        byte_string = input_string if isinstance(input_string, bytes) else str.encode(input_string)
        return base64.b64encode(base64.b64decode(byte_string)) == byte_string
    except Exception:
        return False


class deprecated_keyword_param:
    """A decorator to use to warn when a keyword parameter from a function has been deprecated
    and is intended for future removal. Will emit a warning such a keyword is passed."""

    def __init__(self, keywords, version, reason):
        self.keywords = set(keywords)
        self.version = version
        self.reason = reason

    def __call__(self, fn):
        def wrapper(*args, **kwargs):
            found = self.keywords.intersection(kwargs)
            if found:
                warnings.warn(
                    "Parameter(s) {} deprecated since version {}; {}".format(
                        sorted(list(found)), self.version, self.reason
                    ),
                    category=DeprecationWarning,
                    stacklevel=2
                )

            return fn(*args, **kwargs)

        return wrapper


class Spinner:
    def __init__(self, msg=""):
        self._tick = 0
        self.msg = msg

    def print_tick(self):
        spinner = ['|', '/', '-', '\\'][self._tick % 4]
        if sys.stdin.isatty():
            sys.stdout.write(f"\r {spinner} {self.msg}")
            sys.stdout.flush()
        self._tick += 1
